@@ -3,6 +3,9 @@
 #include "exm-extension.h"
 #include "shell-dbus-interface.h"
 
+#include "../exm-types.h"
+#include "../exm-enums.h"
+
 struct _ExmManager
 {
     GObject parent_instance;
@@ -13,6 +16,8 @@ struct _ExmManager
 
     const gchar *shell_version;
     gboolean extensions_enabled;
+
+    guint update_callback_id;
 };
 
 G_DEFINE_FINAL_TYPE (ExmManager, exm_manager, G_TYPE_OBJECT)
@@ -28,6 +33,14 @@ enum {
 
 static GParamSpec *properties [N_PROPS];
 
+enum {
+    SIGNAL_0,
+    SIGNAL_UPDATES_AVAILABLE,
+    N_SIGNALS
+};
+
+static guint signals [N_SIGNALS];
+
 ExmManager *
 exm_manager_new (void)
 {
@@ -324,6 +337,83 @@ exm_manager_install_finish (ExmManager    *self,
     return g_task_propagate_boolean (G_TASK (result), error);
 }
 
+static int
+list_model_get_number_of_updates (GListModel *model)
+{
+    int num_updates = 0;
+
+    int n_items = g_list_model_get_n_items (model);
+    for (int i = 0; i < n_items; i++)
+    {
+        ExmExtension *ext = g_list_model_get_item (model, i);
+
+        gboolean has_update;
+        g_object_get (ext, "has-update", &has_update, NULL);
+
+        if (has_update)
+            num_updates++;
+    }
+
+    return num_updates;
+}
+
+static guint
+notify_extension_updates (ExmManager *self)
+{
+    // Checks for updates in the background and if found then
+    // emits the 'updates-available' signal with the number of updates
+    int n_updates = 0;
+
+    n_updates += list_model_get_number_of_updates (self->user_ext_model);
+    n_updates += list_model_get_number_of_updates (self->system_ext_model);
+
+    g_info ("There are %d new updates available.", n_updates);
+
+    if (n_updates > 0)
+        g_signal_emit (G_OBJECT (self), signals [SIGNAL_UPDATES_AVAILABLE], 0, n_updates);
+
+    self->update_callback_id = 0;
+
+    return G_SOURCE_REMOVE;
+}
+
+static void
+queue_notify_extension_updates (ExmManager *self)
+{
+    // See: notify_extension_updates
+
+    if (self->update_callback_id != 0)
+        return;
+
+    self->update_callback_id = g_timeout_add (0, G_SOURCE_FUNC (notify_extension_updates), self);
+}
+
+static void
+check_for_updates_done (ShellExtensions *proxy,
+                        GAsyncResult    *res,
+                        ExmManager      *self)
+{
+    GError *error = NULL;
+    shell_extensions_call_check_for_updates_finish (proxy, res, &error);
+
+    if (error)
+    {
+        g_critical ("Could not check for updates: %s", error->message);
+    }
+
+    // Notify the user if updates are detected
+    queue_notify_extension_updates (self);
+}
+
+void
+exm_manager_check_for_updates (ExmManager *self)
+{
+    shell_extensions_call_check_for_updates (self->proxy,
+                                             NULL,
+                                             (GAsyncReadyCallback) check_for_updates_done,
+                                             self);
+}
+
 static void
 exm_manager_class_init (ExmManagerClass *klass)
 {
@@ -362,6 +452,15 @@ exm_manager_class_init (ExmManagerClass *klass)
                                 G_PARAM_READWRITE);
 
     g_object_class_install_properties (object_class, N_PROPS, properties);
+
+    signals [SIGNAL_UPDATES_AVAILABLE]
+        = g_signal_new ("updates-available",
+                        G_TYPE_FROM_CLASS (object_class),
+                        G_SIGNAL_RUN_LAST|G_SIGNAL_NO_RECURSE|G_SIGNAL_NO_HOOKS,
+                        0, NULL, NULL, NULL,
+                        G_TYPE_NONE, 1,
+                        G_TYPE_INT,
+                        NULL);
 }
 
 static void
@@ -378,13 +477,13 @@ parse_single_extension (ExmExtension **extension,
     gchar *uuid = NULL;
     gchar *display_name = NULL;
     gchar *description = NULL;
-    gboolean enabled = FALSE;
     gboolean has_prefs = FALSE;
     gboolean has_update = FALSE;
     gboolean can_change = TRUE;
-
-    *is_user = FALSE;
-    *is_uninstall_operation = FALSE;
+    ExmExtensionState state;
+    ExmExtensionType type;
+    gchar *version = NULL;
+    gchar *error_msg = NULL;
 
     if (extension && *extension)
     {
@@ -413,20 +512,15 @@ parse_single_extension (ExmExtension **extension,
         }
         else if (strcmp (prop_name, "type") == 0)
         {
-            double type;
-            g_variant_get (prop_value, "d", &type);
-            *is_user = (type == 2);
+            gdouble val;
+            g_variant_get (prop_value, "d", &val);
+            type = (ExmExtensionType)val;
         }
         else if (strcmp (prop_name, "state") == 0)
         {
-            double state;
-            g_variant_get (prop_value, "d", &state);
-            enabled = (state == 1);
-
-            if (state == 99)
-            {
-                *is_uninstall_operation = TRUE;
-            }
+            gdouble val;
+            g_variant_get (prop_value, "d", &val);
+            state = (ExmExtensionState)val;
         }
         else if (strcmp (prop_name, "name") == 0)
         {
@@ -448,16 +542,31 @@ parse_single_extension (ExmExtension **extension,
         {
             g_variant_get (prop_value, "b", &can_change);
         }
+        else if (strcmp (prop_name, "version") == 0)
+        {
+            gdouble val;
+            g_variant_get (prop_value, "d", &val);
+            version = g_strdup_printf ("%d", (gint)val);
+        }
+        else if (strcmp (prop_name, "error") == 0)
+        {
+            g_variant_get (prop_value, "s", &error_msg);
+        }
     }
 
+    *is_user = (type == EXM_EXTENSION_TYPE_PER_USER);
+    *is_uninstall_operation = (state == EXM_EXTENSION_STATE_UNINSTALLED);
+
     g_object_set (*extension,
                   "display-name", display_name,
                   "description", description,
-                  "enabled", enabled,
+                  "state", state,
                   "is-user", *is_user,
                   "has-prefs", has_prefs,
                   "has-update", has_update,
                   "can-change", can_change,
+                  "version", version,
+                  "error-msg", error_msg,
                   NULL);
 
     g_free (uuid);
@@ -529,6 +638,8 @@ update_extension_list (ExmManager *self)
 
     g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_USER_EXTENSIONS]);
     g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_SYSTEM_EXTENSIONS]);
+
+    queue_notify_extension_updates (self);
 }
 
 static gboolean
@@ -541,26 +652,6 @@ is_extension_equal (ExmExtension *a, ExmExtension *b)
     return strcmp (uuid_a, uuid_b) == 0;
 }
 
-static gboolean
-replace_extension_in_list_store (GListStore   *store,
-                                 ExmExtension *to_replace)
-{
-    guint position;
-    ExmExtension *to_delete;
-
-    if (g_list_store_find_with_equal_func (store, to_replace, (GEqualFunc)is_extension_equal, &position))
-    {
-        to_delete = g_list_model_get_item (G_LIST_MODEL (store), position);
-        g_list_store_remove (store, position);
-        g_list_store_insert (store, position, to_replace);
-        g_object_unref (to_delete);
-
-        return TRUE;
-    }
-
-    return FALSE;
-}
-
 static void
 on_state_changed (ShellExtensions *object,
                   const gchar     *arg_uuid,
@@ -604,18 +695,23 @@ on_state_changed (ShellExtensions *object,
 
         return;
     }
-}
 
-static void
-on_status_changed (ShellExtensions *object,
-                   const gchar     *arg_uuid,
-                   gint             arg_state,
-                   const gchar     *arg_error,
-                   ExmManager      *self)
-{
-    g_debug ("Status Changed (Unhandled) for extension '%s'\n", arg_uuid);
+    // Emit items-changed signal to re-sort extension list
+    {
+        guint position;
+        if (g_list_store_find_with_equal_func (list_store, extension, (GEqualFunc)is_extension_equal, &position))
+            g_list_model_items_changed (G_LIST_MODEL (list_store), position, 1, 1);
+    }
 
-    // TODO: What's this for?
+    // If the extension that has changed has an update, then
+    // one or more extensions have updates available. Lazily
+    // check the exact number and emit the 'updates-available'
+    // signal.
+    gboolean has_update;
+    g_object_get (extension, "has-update", &has_update, NULL);
+
+    if (has_update)
+        queue_notify_extension_updates (self);
 }
 
 static void
@@ -635,6 +731,8 @@ exm_manager_init (ExmManager *self)
         return;
     }
 
+    self->update_callback_id = 0;
+
     g_object_bind_property (self->proxy, "shell-version",
                             self, "shell-version",
                             G_BINDING_SYNC_CREATE);
@@ -645,6 +743,8 @@ exm_manager_init (ExmManager *self)
 
     update_extension_list (self);
 
-    g_signal_connect (self->proxy, "extension-state-changed", G_CALLBACK (on_state_changed), self);
-    g_signal_connect (self->proxy, "extension-status-changed", G_CALLBACK (on_status_changed), self);
+    g_signal_connect (self->proxy,
+                      "extension-state-changed",
+                      G_CALLBACK (on_state_changed),
+                      self);
 }
